----------------------------------------------------------------------------------
-- Matthew Netsch
--
-- 11 April 2010
--
-- This component provides the approximation of the sigmoid function.
-- Sigmoid(x) = 1/(1+e^(-x))
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

----------------------------------------------------------------------------------
-- Sigmoid
--
-- Provides an approximation for Sigmoid(x) = 1/(1+e^(-x))
-- All floating point values are encoded as a 16-bit integer value
-- where 0xFFFF = 1.0 and 0x0000 = -1.0
--
-- Inputs:
-- 			x     - Input Data Bus
-- 			clk   - Clock 
-- 			ce    - Chip Enable
-- 
-- Outputs:
-- 			oe    - Output Enable
--				y		- Output Data Bus 
--
----------------------------------------------------------------------------------
entity sigmoid is
    Port ( clk 	: in   STD_LOGIC;
           ce     : in   STD_LOGIC;
			  x 		: in   STD_LOGIC_VECTOR (31 downto 0);
           y      : out  STD_LOGIC_VECTOR (15 downto 0);
			  oe     : out  STD_LOGIC);
end sigmoid;

architecture Behavioral of sigmoid is

----------------------------------------------------------------------------------
-- Types
----------------------------------------------------------------------------------
type state is
(
   stIdle,
	stStartApprox,
	stDone,
	stWait
);

component sigLUT
		Port
		(
			clk : in  STD_LOGIC;
			 en : in  STD_LOGIC;
		  addr : in  STD_LOGIC_VECTOR( 14 downto 0 );
		  data : out STD_LOGIC_VECTOR( 15 downto 0 )		
		);
end component;

----------------------------------------------------------------------------------
-- Signals
----------------------------------------------------------------------------------

-- State Machine Sigs
signal stCurrent    : state       := stIdle;
signal stNext       : state       := stIdle;

signal lutSig       : std_logic_vector (15 downto 0);
signal addrSig      : std_logic_vector (14 downto 0);
signal x_sig        : std_logic_vector (31 downto 0);

begin
   
	lut : sigLUT Port Map ( clk => clk,
				                en => '1',
								  addr => addrSig,
								  data => lutSig);
	
   ----------------------------------------------------------------
	-- State Machine Process
	----------------------------------------------------------------
	process(clk)
	begin
	
	  if clk = '1' and clk'event then	  
			stCurrent <= stNext;
	  end if;
	
	end process;
	
	process(clk, ce)
	begin
	
		if ce = '1' and ce'event then
			x_sig <= x;
		end if;
		
	end process;
	
	----------------------------------------------------------------
	-- Next State/Output Logic
	----------------------------------------------------------------
	process(stCurrent, ce)
	begin
	
		case stCurrent is
		
		   when stIdle =>
			   oe <= '0';

				if ce = '1' then
				   stNext <= stStartApprox;
				else
				   stNext <= stIdle;
				end if;
			
			when stStartApprox =>
			   oe <= '0';

				stNext <= stDone;
							
			when stDone =>
			   oe <= '1';
	
				stNext <= stWait;
				
		  when stWait =>
				 oe <= '1';

				 if ce = '0' then
				   stNext <= stIdle;
				 else
				   stNext <= stWait;
				 end if;				
		
		   when others =>
			   oe <= '0';

		      stNext <= stIdle;
		end case;	
	end process;
	
	------------------------------------------------------------------
	-- Approx Process
	------------------------------------------------------------------
	process( clk, ce, x_sig, lutSig )
	
	  variable sub : std_logic_vector(16 downto 0);
	begin
		if clk = '1' and clk'event then
		 if( ce = '1' ) then	 
			 if x_sig >= "00000000010100000000000000000000" then
			 
			    if x_sig <= "01111111111111111111111111111111" then
				   -- Greater than 5 					
				   y <= "1111111111111111";
					addrSig <= "000000000000000";
				 
				 elsif x <= "11111111101100000000000000000000" then
				   -- Less than -5					
					y <= "0000000000000000";
					addrSig <= "000000000000000";
		 
				 else
				   -- Less than 0 but greater than -5
               sub := "01111111111111111" - ("0" & lutSig);
					addrSig <= not(x_sig(22 downto 8)) + 1;
               
				   y <= sub(15 downto 0);
				 end if;
				 
			 else
			      -- Less than 5 but greater than 0
					addrSig <= x_sig(22 downto 8);
       		   y <= lutSig;
			 end if;
	--	 else
	--	    y <= "0000000000000000";
	--		 addrSig <= "000000000000000";
	  end if;
	 end if;
		
	end process;

end Behavioral;

